from abc import (
    ABC,
    abstractmethod,
)
from datetime import timedelta
from typing import Any

import attrs
import testcontainers_yt_local.container
from yt.wrapper.client import YtClient
from yt.wrapper.default_config import retries_config


class YtInstance(ABC):
    @abstractmethod
    def get_client(self) -> YtClient:
        pass


@attrs.define
class YtInstanceExternal(YtInstance):
    proxy_url: str
    token: str = attrs.field(repr=False)

    def get_client(self) -> YtClient:
        return YtClient(
            proxy=self.proxy_url,
            token=self.token,
        )


class YtInstanceTestContainers(YtInstance):
    def __init__(self) -> None:
        self.yt_container = testcontainers_yt_local.container.YtLocalContainer(
            use_ng_image=True,
            enable_cri_jobs=True,
            privileged=True,
            set_yt_config_env_var=True,
        )

    def __enter__(self) -> "YtInstanceTestContainers":
        self.yt_container.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.yt_container.stop()

    def get_client(self) -> YtClient:
        return self.yt_container.get_client(
            config={
                "proxy": {
                    "retries": retries_config(
                        enable=True,
                        total_timeout=timedelta(minutes=5),
                        backoff={
                            "policy": "constant_time",
                            "constant_time": 5,
                        },
                    ),
                },
            },
        )



#!/usr/bin/env python3
import os
from pathlib import Path
import sys

import cattrs

from tractorun.exception import TractorunVersionMismatchError
from tractorun.private.bind import BindsPacker
from tractorun.private.bootstrapper import (
    BootstrapConfig,
    bootstrap,
)
from tractorun.private.constants import (
    BIND_PATHS_ENV_VAR,
    BOOTSTRAP_CONFIG_FILENAME_ENV_VAR,
    JOB_SANDBOX_PATH,
)
from tractorun.private.helpers import (
    AttrSerializer,
    create_attrs_converter,
)


def main() -> None:
    binds_packer = BindsPacker.from_env(os.environ[BIND_PATHS_ENV_VAR])
    binds_packer.unpack()
    bootstrap_config_path = os.environ[BOOTSTRAP_CONFIG_FILENAME_ENV_VAR]
    with open(bootstrap_config_path, "r") as f:
        content = f.read()
        deserializer = AttrSerializer(
            BootstrapConfig,
            # forward compatibility
            converter=create_attrs_converter(forbid_extra_keys=False),
        )
        try:
            config: BootstrapConfig = deserializer.deserialize(data=content)
        except cattrs.errors.BaseValidationError as e:
            raise TractorunVersionMismatchError(
                "Please check that the tractorun version locally and on YT are the same",
            ) from e
    bootstrap(
        mesh=config.mesh,
        training_dir=config.training_dir,
        yt_client_config=config.yt_client_config,
        sidecars=config.sidecars,
        env=config.env,
        command=sys.argv[1:],
        tensorproxy=config.tensorproxy,
        lib_versions=config.lib_versions,
        cluster_config=config.cluster_config,
        operation_log_mode=config.operation_log_mode,
        sandbox_path=Path(JOB_SANDBOX_PATH),
    )


if __name__ == "__main__":
    main()



import pytest
from yt.common import YT_NULL_TRANSACTION_ID

from tests.utils import (
    GENERIC_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.run import run


def test_configuration() -> None:
    _, _, config = make_configuration(make_cli_args())
    assert config.no_wait is False

    run_config = make_run_config({})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.no_wait is False

    run_config = make_run_config({"no_wait": False})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path, "--no-wait"])
    assert config.no_wait is True

    run_config = make_run_config({"no_wait": True})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.no_wait is True


@pytest.mark.parametrize(
    "no_wait,expected",
    [
        (True, True),
        (False, False),
    ],
)
def test_transaction_pickle(no_wait: bool, expected: bool, yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    def dummy(toolbox: TractoCli) -> None:
        pass

    run_info = run(
        dummy,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
        no_wait=no_wait,
    )
    operation_info = yt_client.get_operation(run_info.operation_id)
    assert (operation_info["user_transaction_id"] == YT_NULL_TRANSACTION_ID) is expected


@pytest.mark.parametrize(
    "no_wait,expected",
    [
        (True, True),
        (False, False),
    ],
)
def test_transaction_script(no_wait: bool, expected: bool, yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()
    args = [
        "--yt-path",
        yt_path,
        "--bind-local",
        f"{get_data_path('../data/dummy_script.py')}:/tractorun_tests/dummy_script.py",
    ]
    if no_wait:
        args.append("--no-wait")

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/dummy_script.py"],
        args=args,
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    operation_info = op_run.get_operation_info(yt_client)
    assert (operation_info["user_transaction_id"] == YT_NULL_TRANSACTION_ID) is expected



import contextlib
import inspect
import json
from pathlib import Path
import random
import string
import subprocess
import tempfile
import time
from types import ModuleType
from typing import (
    Any,
    Generator,
    Iterable,
)
import uuid

import attrs
import yaml
from yt import wrapper as yt


_CURRENT_DOCKER_TAG = "2025-01-15-20-21-21"
TRACTORCH_DOCKER_IMAGE: str = f"ghcr.io/tractoai/tractorun-tractorch-tests:{_CURRENT_DOCKER_TAG}"
TRACTORAX_DOCKER_IMAGE: str = f"ghcr.io/tractoai/tractorun-tractorax-tests:{_CURRENT_DOCKER_TAG}"
GENERIC_DOCKER_IMAGE: str = f"ghcr.io/tractoai/tractorun-generic-tests:{_CURRENT_DOCKER_TAG}"
EXAMPLES_DOCKER_IMAGE: str = f"ghcr.io/tractoai/tractorun-examples-runtime:{_CURRENT_DOCKER_TAG}"
TENSORPROXY_DOCKER_IMAGE: str = f"ghcr.io/tractoai/tractorun-tensorproxy-tests:{_CURRENT_DOCKER_TAG}"

OPERATION_INFO_RETRIES = 5
OPERATION_INFO_TIMEOUT = 2


def get_data_path(filename: str | Path) -> Path:
    return (Path(__file__).parent / "data" / filename).resolve()


def get_random_string(length: int) -> str:
    return str(int(time.time())) + "_" + "".join(random.choice(string.ascii_letters) for _ in range(length))


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TractoCliRun:
    _process: subprocess.Popen
    _operation_title: str

    def is_exitcode_valid(self, exit_code: int = 0) -> bool:
        return self._process.returncode == exit_code

    def get_operation_info(self, yt_client: yt.YtClient) -> dict:
        operations = []
        for _ in range(OPERATION_INFO_RETRIES):
            operations = yt_client.list_operations(filter=self._operation_title)["operations"]
            if operations:
                break
            time.sleep(OPERATION_INFO_TIMEOUT)

        assert len(operations) == 1

        operation_id = operations[0]["id"]
        operation_info = yt_client.get_operation(operation_id)
        return operation_info

    def is_operation_state_valid(self, yt_client: yt.YtClient, job_count: int) -> bool:
        operation_info = self.get_operation_info(yt_client)
        operation_spec = operation_info["spec"]
        return operation_spec["tasks"]["task"]["job_count"] == job_count

    @property
    def stdout(self) -> bytes:
        assert self._process.stdout is not None
        data = self._process.stdout.read()
        return data

    @property
    def stderr(self) -> bytes:
        assert self._process.stderr is not None
        data = self._process.stderr.read()
        return data


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TractoCli:
    _command: list[str | Path]
    _docker_image: str | None = attrs.field(default=GENERIC_DOCKER_IMAGE)
    _args: list[str]
    _task_spec: dict[str, Any] = attrs.field(default={})
    _operation_spec: dict[str, Any] = attrs.field(default={})

    def run(self) -> TractoCliRun:
        command, operation_title, task_title = self._prepare_command()

        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        process.wait()
        return TractoCliRun(
            process=process,
            operation_title=operation_title,
        )

    def dry_run(self) -> dict:
        command, operation_title, task_title = self._prepare_command()
        command.append("--dry-run")

        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        process.wait()
        run = TractoCliRun(
            process=process,
            operation_title=operation_title,
        )
        assert run.is_exitcode_valid()
        return json.loads(run.stdout)

    def _prepare_command(self) -> tuple[list[str | Path], str, str]:
        operation_title = f"test operation {uuid.uuid4()}"
        task_title = f"test operation's task {uuid.uuid4()}"

        def _get_docker_image_arg() -> tuple:
            if self._docker_image is not None:
                return ("--docker-image", self._docker_image)
            return tuple()

        command: list[str | Path] = [
            str(get_data_path("../../tractorun/cli/tractorun_runner.py")),
            *_get_docker_image_arg(),
            "--title",
            operation_title,
            "--yt-operation-spec",
            json.dumps(
                {
                    **self._operation_spec,
                },
            ),
            "--yt-task-spec",
            json.dumps(
                {
                    "title": task_title,
                    **self._task_spec,
                },
            ),
            "--bind-local-lib",
            str(get_data_path("../../tractorun")),
            *self._args,
            *self._command,
        ]
        return command, operation_title, task_title


@contextlib.contextmanager
def run_config_file(config: dict[str, Any]) -> Generator[str, None, None]:
    with tempfile.NamedTemporaryFile(mode="w") as f:
        yaml.safe_dump(config, f)
        yield f.name


def make_cli_args(*cli_args: str) -> list:
    return ["--yt-path", "foo", "--docker-image", GENERIC_DOCKER_IMAGE, *cli_args, "command"]


def make_run_config(config: dict[str, Any]) -> dict[str, Any]:
    return {
        "command": ["foo"],
        "yt_path": "foo",
        "docker_image": GENERIC_DOCKER_IMAGE,
        **config,
    }


def _is_private_module(module: ModuleType | None) -> bool:
    if module is None:
        return False
    private_prefixes = [
        "tractorun.private",
        "tractorun.tests",
        "tractorun.cli",
    ]
    for prefix in private_prefixes:
        if module.__name__.startswith(prefix):
            return True
    return False


def check_no_private_objects_in_public_namespace(modules: Iterable[ModuleType]) -> bool:
    # this test can't check variables with primitive types like strings, int, float and so on
    invalid_public_modules: list[tuple[str, object, ModuleType]] = []
    for module in modules:
        if _is_private_module(module):
            continue
        for name, obj in inspect.getmembers(module):
            if name.startswith("_"):
                continue
            if not isinstance(obj, object):
                continue
            if inspect.ismodule(obj):
                continue
            obj_module = inspect.getmodule(obj)
            if obj_module is None:
                continue
            if not _is_private_module(obj_module):
                continue
            invalid_public_modules.append((name, obj, module))
    return invalid_public_modules == []



import json
import sys
from typing import (
    Any,
    Callable,
)

import pytest
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data

from tests.utils import (
    TRACTORCH_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
)
from tests.yt_instances import YtInstance
from tractorun.backend.tractorch import Tractorch
from tractorun.backend.tractorch.dataset import YtTensorDataset
from tractorun.backend.tractorch.serializer import TensorSerializer
from tractorun.mesh import Mesh
from tractorun.run import run
from tractorun.toolbox import Toolbox


def _get_simple_train(mnist_ds_path: str) -> Callable:
    def _simple_train(toolbox: Toolbox) -> None:
        class Net(nn.Module):
            def __init__(self) -> None:
                super(Net, self).__init__()
                self.l1 = torch.nn.Linear(28 * 28, 10)

            def forward(self, x: Any) -> torch.Tensor:
                return torch.relu(self.l1(x.view(x.size(0), -1)))

        device = torch.device("cpu")
        serializer = TensorSerializer()
        train_dataset = YtTensorDataset(toolbox.yt_client, mnist_ds_path)
        train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=64)
        model = Net().to(device)
        optimizer = optim.Adadelta(model.parameters(), lr=1.0)
        model.train()
        final_loss = None
        for batch_idx, (data, target) in enumerate(train_loader):
            data, target = data.to(device), target.to(device)
            optimizer.zero_grad()
            output = model(data)
            loss = F.nll_loss(output, target)
            loss.backward()
            optimizer.step()
            final_loss = loss.item()
        toolbox.save_model(
            data=serializer.serialize(model.state_dict()),
            dataset_path=mnist_ds_path,
            metadata={
                "loss": str(final_loss),
            },
        )

    return _simple_train


def test_run_pickle(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    train_func = _get_simple_train(mnist_ds_path)

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    # The operation did not fail => success!
    run(
        train_func,
        backend=Tractorch(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=TRACTORCH_DOCKER_IMAGE,
    )


def test_run_torch_distributed(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    train_func = _get_simple_train(mnist_ds_path)

    mesh = Mesh(node_count=2, process_per_node=1, gpu_per_process=0)
    # The operation did not fail => success!
    run(
        train_func,
        backend=Tractorch(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=TRACTORCH_DOCKER_IMAGE,
    )


def test_run_torch_with_checkpoints(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    def train(toolbox: Toolbox) -> None:
        class Net(nn.Module):
            def __init__(self) -> None:
                super(Net, self).__init__()
                self.l1 = torch.nn.Linear(28 * 28, 10)

            def forward(self, x: Any) -> torch.Tensor:
                return torch.relu(self.l1(x.view(x.size(0), -1)))

        device = torch.device("cpu")
        serializer = TensorSerializer()
        model = Net().to(device)
        optimizer = optim.Adadelta(model.parameters(), lr=1.0)

        checkpoint = toolbox.checkpoint_manager.get_last_checkpoint()
        first_batch_index = 0
        if checkpoint is not None:
            first_batch_index = checkpoint.metadata["first_batch_index"]
            print(
                "Found checkpoint with index",
                checkpoint.index,
                "and first batch index",
                first_batch_index,
                file=sys.stderr,
            )

            checkpoint_dict = serializer.desirialize(checkpoint.value)
            model.load_state_dict(checkpoint_dict["model"])
            optimizer.load_state_dict(checkpoint_dict["optimizer"])

        train_dataset = YtTensorDataset(toolbox.yt_client, mnist_ds_path)
        train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=10)

        model.train()
        for batch_idx, (data, target) in enumerate(train_loader):
            # TODO: Do it normally. YT dataloader?
            if batch_idx < first_batch_index:
                continue

            if batch_idx == 3 and first_batch_index == 0:
                raise Exception("Force restart to test checkpoints")

            data, target = data.to(device), target.to(device)
            optimizer.zero_grad()
            output = model(data)
            loss = F.nll_loss(output, target)
            loss.backward()
            optimizer.step()

            if batch_idx % 3 == 0:
                state_dict = {
                    "model": model.state_dict(),
                    "optimizer": optimizer.state_dict(),
                }
                metadata_dict = {
                    "first_batch_index": batch_idx + 1,
                    "loss": loss.item(),
                }
                toolbox.checkpoint_manager.save_checkpoint(serializer.serialize(state_dict), metadata_dict)
                print("Saved checkpoint after batch with index", batch_idx, file=sys.stderr)

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)

    def _run() -> None:
        run(
            train,
            backend=Tractorch(),
            yt_path=yt_path,
            mesh=mesh,
            yt_client=yt_client,
            docker_image=TRACTORCH_DOCKER_IMAGE,
        )

    # First launch is failed because of the exception in the train function.
    with pytest.raises(Exception):
        _run()
    # And second one should be successful.
    _run()


def test_run_script(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/torch_run_script.py"],
        docker_image=TRACTORCH_DOCKER_IMAGE,
        args=[
            "--mesh.node-count",
            "1",
            "--mesh.process-per-node",
            "1",
            "--mesh.gpu-per-process",
            "0",
            "--yt-path",
            yt_path,
            "--user-config",
            json.dumps({"MNIST_DS_PATH": mnist_ds_path}),
            "--bind-local",
            f"{get_data_path('../data/torch_run_script.py')}:/tractorun_tests/torch_run_script.py",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)



import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms
from torch.optim.lr_scheduler import StepLR

from tractorun.backend.tractorch import Tractorch
from tractorun.backend.tractorch.serializer import TensorSerializer
from tractorun.run import prepare_and_get_toolbox

class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


def train(args, model, device, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), loss.item()))
            if args.dry_run:
                break


def test(model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
        test_loss, correct, len(test_loader.dataset),
        100. * correct / len(test_loader.dataset)))


def main():
    toolbox = prepare_and_get_toolbox(backend=Tractorch())

    # Training settings
    parser = argparse.ArgumentParser(description='PyTorch MNIST Example')
    parser.add_argument('--batch-size', type=int, default=64, metavar='N',
                        help='input batch size for training (default: 64)')
    parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--epochs', type=int, default=14, metavar='N',
                        help='number of epochs to train (default: 14)')
    parser.add_argument('--lr', type=float, default=1.0, metavar='LR',
                        help='learning rate (default: 1.0)')
    parser.add_argument('--gamma', type=float, default=0.7, metavar='M',
                        help='Learning rate step gamma (default: 0.7)')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--no-mps', action='store_true', default=False,
                        help='disables macOS GPU training')
    parser.add_argument('--dry-run', action='store_true', default=False,
                        help='quickly check a single pass')
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                        help='how many batches to wait before logging training status')
    parser.add_argument('--save-model', action='store_true', default=False,
                        help='For Saving the current Model')
    args = parser.parse_args([])
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    use_mps = not args.no_mps and torch.backends.mps.is_available()

    torch.manual_seed(args.seed)

    if use_cuda:
        device = torch.device("cuda")
    elif use_mps:
        device = torch.device("mps")
    else:
        device = torch.device("cpu")

    train_kwargs = {'batch_size': args.batch_size}
    test_kwargs = {'batch_size': args.test_batch_size}
    if use_cuda:
        cuda_kwargs = {'num_workers': 1,
                       'pin_memory': True,
                       'shuffle': True}
        train_kwargs.update(cuda_kwargs)
        test_kwargs.update(cuda_kwargs)

    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])
    dataset1 = datasets.MNIST('../data', train=True, download=True,
                              transform=transform)
    dataset2 = datasets.MNIST('../data', train=False,
                              transform=transform)

    train_loader = torch.utils.data.DataLoader(dataset1,**train_kwargs)
    test_loader = torch.utils.data.DataLoader(dataset2, **test_kwargs)

    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=args.lr)

    scheduler = StepLR(optimizer, step_size=1, gamma=args.gamma)
    for epoch in range(1, args.epochs + 1):
        train(args, model, device, train_loader, optimizer, epoch)
        test(model, device, test_loader)
        scheduler.step()

    if args.save_model:
        ts = TensorSerializer()
        toolbox.save_model(ts.serialize(model.state_dict()))


if __name__ == "__main__":
    main()



from tractorun.backend.generic import GenericBackend
from tractorun.run import prepare_and_get_toolbox


def main() -> None:
    _ = prepare_and_get_toolbox(backend=GenericBackend())


if __name__ == "__main__":
    main()



from typing import Any
from urllib.parse import quote
import warnings

import attrs
import yt.wrapper as yt


def _to_str(value: Any) -> str | None:
    if value is None:
        return None
    return str(value)


def make_cypress_link(cypress_link_template: str | None, path: str) -> str | None:
    if cypress_link_template is None:
        return None
    return cypress_link_template.format(path=quote(path))


def make_job_stderr_link(job_stderr_link_template: str | None, operation_id: str, job_id: str) -> str | None:
    if job_stderr_link_template is None:
        return None
    return job_stderr_link_template.format(
        job_id=quote(job_id),
        operation_id=quote(operation_id),
    )


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TractorunClusterConfig:
    cypress_link_template: str | None
    job_stderr_link_template: str | None

    @staticmethod
    def load_from_yt(yt_client: yt.YtClient, path: str, fail_if_not_exist: bool = False) -> "TractorunClusterConfig":
        if not fail_if_not_exist and not yt_client.exists(path):
            warnings.warn(
                f"Cluster config {path} does not exist. Some functions are not available. Please specify config's path by tractorun params.",
            )
            return TractorunClusterConfig(
                cypress_link_template=None,
                job_stderr_link_template=None,
            )

        config = yt_client.get(path)
        if not config:
            config = {}

        return TractorunClusterConfig(
            cypress_link_template=_to_str(config.get("cypress_link_template")),
            job_stderr_link_template=_to_str(config.get("job_stderr_link_template")),
        )

    def to_dict(self) -> dict:
        return attrs.asdict(self)  # type: ignore



import pathlib

import attrs


def _to_abs_local(path: str) -> str:
    return str(pathlib.Path(path).absolute())


def _to_abs_posix(path: str) -> str:
    return str(pathlib.PosixPath(path).absolute())


__all__ = ["BindLocal", "BindCypress", "BindAttributes"]


@attrs.define(kw_only=True, slots=True)
class BindAttributes:
    executable: bool = attrs.field(default=True)
    format: str | None = attrs.field(default=None)
    bypass_artifact_cache: bool = attrs.field(default=False)


@attrs.define(kw_only=True, slots=True)
class BindLocal:
    # TODO: just use pathlib
    source: str = attrs.field(converter=_to_abs_local)
    destination: str = attrs.field(converter=_to_abs_posix)


@attrs.define(kw_only=True, slots=True)
class BindCypress:
    source: str = attrs.field()
    destination: str = attrs.field()
    attributes: BindAttributes = attrs.field(default=BindAttributes())



import socket
import sys
from urllib.parse import urlparse

import jax

from tractorun.base_backend import EnvironmentBase
from tractorun.private.closet import Closet as _Closet


__all__ = ["Environment"]


class Environment(EnvironmentBase):
    def prepare(self, closet: _Closet) -> None:
        local_device_ids = None
        if closet.mesh.gpu_per_process > 0:
            process_id = closet.coordinator.get_process_index()
            first_device_index = process_id * closet.mesh.gpu_per_process
            local_device_ids = list(range(first_device_index, first_device_index + closet.mesh.gpu_per_process))

        coordinator_address = closet.coordinator.get_primary_endpoint()
        print("old coordinator address: {}".format(coordinator_address), file=sys.stderr)
        parsed_coordinator_address = urlparse(f"schema://{coordinator_address}")
        # because of overlay problems
        if parsed_coordinator_address.hostname == socket.gethostname():
            coordinator_address = f"127.0.0.1:{parsed_coordinator_address.port}"
            print("new coordinator address: {}".format(coordinator_address), file=sys.stderr)

        jax.distributed.initialize(
            coordinator_address=coordinator_address,
            num_processes=closet.coordinator.get_total_peer_count(),
            process_id=closet.coordinator.get_self_index(),
            local_device_ids=local_device_ids,
        )



from tests.utils import (
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.cli.tractorun_runner import (
    CLUSTER_CONFIG_PATH_DEFAULT,
    make_configuration,
)
from tractorun.private.yt_cluster import (
    TractorunClusterConfig,
    make_cypress_link,
    make_job_stderr_link,
)


def test_load_config(cluster_config_path: str, cluster_config: TractorunClusterConfig, yt_instance: YtInstance) -> None:
    config = TractorunClusterConfig.load_from_yt(yt_client=yt_instance.get_client(), path=cluster_config_path)
    assert config == cluster_config


def test_configuration() -> None:
    _, _, config = make_configuration(make_cli_args())
    assert config.cluster_config_path == CLUSTER_CONFIG_PATH_DEFAULT

    _, _, config = make_configuration(make_cli_args("--cluster-config-path", "//cli_path"))
    assert config.cluster_config_path == "//cli_path"

    run_config = make_run_config(
        {
            "cluster_config_path": "//config_path",
        }
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.cluster_config_path == "//config_path"

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path, "--cluster-config-path", "//cli_path"])
    assert config.cluster_config_path == "//cli_path"


def test_make_cypress_link(
    cluster_config_path: str, cluster_config: TractorunClusterConfig, yt_instance: YtInstance
) -> None:
    config = TractorunClusterConfig.load_from_yt(yt_client=yt_instance.get_client(), path=cluster_config_path)
    assert (
        make_cypress_link(path="//some/path 1/2", cypress_link_template=config.cypress_link_template)
        == "https://yt.tracto.ai/yt/navigation?path=//some/path%201/2"
    )
    assert make_cypress_link(path="//some/path", cypress_link_template=None) is None


def test_make_job_stderr_link(
    cluster_config_path: str, cluster_config: TractorunClusterConfig, yt_instance: YtInstance
) -> None:
    config = TractorunClusterConfig.load_from_yt(yt_client=yt_instance.get_client(), path=cluster_config_path)
    assert (
        make_job_stderr_link(
            job_id="123-123", operation_id="321-321", job_stderr_link_template=config.job_stderr_link_template
        )
        == "https://yt.tracto.ai/api/yt/yt/api/v3/get_job_stderr?operation_id=321-321&job_id=123-123&dump_error_into_response=true"
    )
    assert (
        make_job_stderr_link(
            job_id="123-123",
            operation_id="321-321",
            job_stderr_link_template=None,
        )
        is None
    )



import socket
import sys
from urllib.parse import urlparse

import torch
import torch.distributed

from tractorun.base_backend import EnvironmentBase
from tractorun.exception import TractorunConfigurationError
from tractorun.private.closet import Closet as _Closet


__all__ = ["Environment"]


class Environment(EnvironmentBase):
    def prepare(self, closet: _Closet) -> None:
        if closet.mesh.gpu_per_process > 0:
            assert torch.cuda.is_available()

            if closet.mesh.gpu_per_process > 1:
                raise TractorunConfigurationError("gpu per process should be <= 1 for tractorch")

            device_index = closet.coordinator.get_process_index()
            assert device_index < torch.cuda.device_count()
            torch.cuda.set_device(closet.coordinator.get_process_index())

        coordinator_address = closet.coordinator.get_primary_endpoint()
        print("old coordinator address: {}".format(coordinator_address), file=sys.stderr)
        parsed_coordinator_address = urlparse(f"schema://{coordinator_address}")
        # because of overlay problems
        if parsed_coordinator_address.hostname == socket.gethostname():
            coordinator_address = f"127.0.0.1:{parsed_coordinator_address.port}"
            print("new coordinator address: {}".format(coordinator_address), file=sys.stderr)

        backend = "gloo" if closet.mesh.gpu_per_process == 0 else "nccl"
        torch.distributed.init_process_group(
            backend=backend,
            init_method="tcp://" + coordinator_address,
            rank=closet.coordinator.get_self_index(),
            world_size=closet.coordinator.get_total_peer_count(),
        )



import base64
import json
import os
import pickle
import socket

import attrs
from yt.wrapper import YtClient

from tractorun.checkpoint import CheckpointManager
from tractorun.coordinator import Coordinator
from tractorun.description import DescriptionManager
from tractorun.mesh import Mesh
from tractorun.private.constants import TRACTO_CONFIG_ENV_VAR
from tractorun.private.coordinator import CoordinatorFactory
from tractorun.private.helpers import AttrSerializer
from tractorun.private.training_dir import TrainingDir
from tractorun.private.worker import WorkerConfig
from tractorun.private.yt_cluster import TractorunClusterConfig


@attrs.define
class TrainingMetadata:
    # TODO: choose a good place for this data
    operation_id: str
    job_id: str


@attrs.define
class Closet:
    mesh: Mesh
    coordinator: Coordinator
    yt_client: YtClient
    checkpoint_manager: CheckpointManager
    training_dir: TrainingDir
    training_metadata: TrainingMetadata
    cluster_config: TractorunClusterConfig
    description_manager: DescriptionManager


def get_closet() -> Closet:
    config_path = os.environ[TRACTO_CONFIG_ENV_VAR]
    with open(config_path, "r") as ff:
        deserializer = AttrSerializer(WorkerConfig)
        config: WorkerConfig = deserializer.deserialize(json.load(ff))

    self_endpoint = socket.gethostname() + ":" + str(config.port)
    yt_client = YtClient(
        config=pickle.loads(
            base64.b64decode(config.yt_client_config),
        ),
    )
    training_metadata = TrainingMetadata(
        operation_id=config.operation_id,
        job_id=config.job_id,
    )
    description_manager = DescriptionManager(
        operation_id=training_metadata.operation_id,
        yt_client=yt_client,
        cypress_link_template=config.cluster_config.cypress_link_template,
    )
    coordinator = CoordinatorFactory(
        yt_client=yt_client,
        training_dir=config.training_dir,
        self_endpoint=self_endpoint,
        mesh=config.mesh,
        process_index=config.proc_index,
        node_index=config.node_index,
        self_index=config.self_index,
        operation_id=training_metadata.operation_id,
        job_id=training_metadata.job_id,
    ).create()

    checkpoint_manager = CheckpointManager(config.training_dir.checkpoints_path, yt_client)
    # TODO: make CheckpointFactory instead of the mystical initialize method
    checkpoint_manager.initialize()

    return Closet(
        mesh=config.mesh,
        coordinator=coordinator,
        yt_client=yt_client,
        training_dir=config.training_dir,
        training_metadata=training_metadata,
        checkpoint_manager=checkpoint_manager,
        cluster_config=config.cluster_config,
        description_manager=description_manager,
    )



import argparse
import os
from pathlib import Path
import random
import string
import sys

from jax import (
    grad,
    jit,
)

from tractorun.backend.tractorax import Tractorax
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


def task(_: Toolbox) -> None:
    @jit
    def f(x: int) -> int:
        return x**2 + 3 * x + 1

    grad_f = grad(f)
    print(grad_f(1.0), file=sys.stderr)


def _default_home_dir() -> str:
    rnm = "".join(random.choices(string.ascii_letters + string.digits, k=8))
    return f"//tmp/tractorun_examples/{rnm}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--yt-home-dir", type=str, default=_default_home_dir())
    parser.add_argument("--pool-tree", type=str, default="default")
    parser.add_argument("--dataset-path", type=str, default="//home/samples/mnist-torch-train")
    parser.add_argument("--docker-image", type=str, default=os.environ.get("DOCKER_IMAGE"))
    parser.add_argument("--gpu-per-process", type=int, default=0)
    args = parser.parse_args()

    mesh = Mesh(node_count=1, process_per_node=8, gpu_per_process=args.gpu_per_process, pool_trees=[args.pool_tree])

    tractorun_path = (Path(__file__).parent.parent.parent.parent / "tractorun").resolve()
    run(
        task,
        backend=Tractorax(),
        yt_path=args.yt_home_dir,
        mesh=mesh,
        docker_image=args.docker_image,
        binds_local_lib=[str(tractorun_path)],
        resources=Resources(cpu_limit=4.0, memory_limit=16 * 1024**3),
    )



import argparse
import os
from pathlib import Path
import random
import string
import sys

import jax

from tractorun.backend.tractorax import Tractorax
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


def task(_: Toolbox) -> None:
    print("Total device count:", jax.device_count(), file=sys.stderr)
    print("Local device count:", jax.local_device_count(), file=sys.stderr)

    xs = jax.numpy.ones(jax.local_device_count())
    print(jax.pmap(lambda x: jax.lax.psum(x, "i"), axis_name="i")(xs), file=sys.stderr)


def _default_home_dir() -> str:
    rnm = "".join(random.choices(string.ascii_letters + string.digits, k=8))
    return f"//tmp/tractorun_examples/{rnm}"


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--yt-home-dir", type=str, default=_default_home_dir())
    parser.add_argument("--pool-tree", type=str, default="default")
    parser.add_argument("--dataset-path", type=str, default="//home/samples/mnist-torch-train")
    parser.add_argument("--docker-image", type=str, default=os.environ.get("DOCKER_IMAGE"))
    parser.add_argument("--gpu-per-process", type=int, default=0)
    args = parser.parse_args()

    mesh = Mesh(node_count=1, process_per_node=8, gpu_per_process=args.gpu_per_process, pool_trees=[args.pool_tree])
    tractorun_path = (Path(__file__).parent.parent.parent.parent / "tractorun").resolve()
    run(
        task,
        backend=Tractorax(),
        yt_path=args.yt_home_dir,
        mesh=mesh,
        docker_image=args.docker_image,
        resources=Resources(cpu_limit=1.0, memory_limit=8076021002),
        binds_local_lib=[str(tractorun_path)],
    )



import sys

from tractorun.backend.generic import GenericBackend
from tractorun.run import prepare_and_get_toolbox


def main() -> None:
    toolbox = prepare_and_get_toolbox(backend=GenericBackend())
    test_strings = toolbox.get_user_config()["test_strings"]
    for test_string in test_strings:
        print(test_string, file=sys.stderr)


if __name__ == "__main__":
    main()



import json
import os
import uuid

from _pytest.monkeypatch import MonkeyPatch
import pytest

from tests.utils import (
    GENERIC_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.bind import BindLocal
from tractorun.cli.tractorun_runner import CliRunInfo
from tractorun.env import EnvVariable
from tractorun.exception import TractorunConfigurationError
from tractorun.mesh import Mesh
from tractorun.private.helpers import AttrSerializer
from tractorun.run import (
    run,
    run_script,
)
from tractorun.toolbox import Toolbox


def test_important_spec_options(yt_path: str) -> None:
    def checker() -> None:
        pass

    run_info = run(
        checker,
        yt_path=yt_path,
        docker_image=GENERIC_DOCKER_IMAGE,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        backend=GenericBackend(),
        dry_run=True,
    )
    assert run_info.operation_spec["annotations"]["is_tractorun"] is True
    assert run_info.operation_spec["fail_on_job_restart"] is True
    assert run_info.operation_spec["is_gang"] is True


def test_prepare_dataset(yt_instance: YtInstance, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()
    row_count = yt_client.get_attribute(mnist_ds_path, "row_count")
    assert row_count == 100
    schema = yt_client.get_attribute(mnist_ds_path, "schema")
    assert len(schema) == 2


def test_run(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    def checker(toolbox: Toolbox) -> None:
        pass

    yt_client = yt_instance.get_client()

    operation_title = f"test operation {uuid.uuid4()}"
    task_title = f"test operation's task {uuid.uuid4()}"

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)

    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
        title=operation_title,
        yt_task_spec={"title": task_title},
    )

    operations = yt_client.list_operations(filter=operation_title)["operations"]
    assert len(operations) == 1

    operation_id = operations[0]["id"]

    operation_spec = yt_client.get_operation(operation_id)["spec"]
    assert operation_spec["title"] == operation_title
    assert operation_spec["tasks"]["task"]["title"] == task_title


def test_run_script(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/dummy_script.py"],
        args=[
            "--mesh.node-count",
            "1",
            "--mesh.process-per-node",
            "1",
            "--mesh.gpu-per-process",
            "0",
            "--yt-path",
            yt_path,
            "--bind-local",
            f"{get_data_path('../data/dummy_script.py')}:/tractorun_tests/dummy_script.py",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)


def test_run_script_with_config(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/dummy_script.py"],
        args=[
            "--run-config-path",
            str(get_data_path("../data/run_config.yaml")),
            "--yt-path",
            yt_path,
            "--user-config",
            json.dumps({"MNIST_DS_PATH": mnist_ds_path}),
            "--bind-local",
            f"{get_data_path('../data/dummy_script.py')}:/tractorun_tests/dummy_script.py",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=2)


@pytest.mark.parametrize(
    "env,expected",
    [
        ({"YT_BASE_LAYER": "custom_image_1"}, "custom_image_1"),
        ({"YT_JOB_DOCKER_IMAGE": "custom_image_2"}, "custom_image_2"),
    ],
)
def test_docker_image_script(yt_path: str, env: dict[str, str], expected: str, monkeypatch: MonkeyPatch) -> None:
    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/dummy_script.py"],
        args=[
            "--mesh.node-count",
            "1",
            "--mesh.process-per-node",
            "1",
            "--mesh.gpu-per-process",
            "0",
            "--yt-path",
            yt_path,
            "--bind-local",
            f"{get_data_path('../data/dummy_script.py')}:/tractorun_tests/dummy_script.py",
        ],
        docker_image=None,
    )
    for key, value in env.items():
        monkeypatch.setenv(key, value)
    run_info = tracto_cli.dry_run()
    assert run_info["configuration"]["effective_config"]["docker_image"] == expected
    assert run_info["run_info"]["operation_spec"]["tasks"]["task"]["docker_image"] == expected


@pytest.mark.parametrize(
    "env,expected",
    [
        ({"YT_BASE_LAYER": "custom_image_1"}, "custom_image_1"),
        ({"YT_JOB_DOCKER_IMAGE": "custom_image_2"}, "custom_image_2"),
    ],
)
def test_docker_image_pickle(yt_path: str, env: dict[str, str], expected: str, monkeypatch: MonkeyPatch) -> None:
    def checker() -> None:
        pass

    for key, value in env.items():
        monkeypatch.setenv(key, value)
    run_info = run(
        checker,
        yt_path=yt_path,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        backend=GenericBackend(),
        dry_run=True,
    )
    assert run_info.operation_spec["tasks"]["task"]["docker_image"] == expected


def test_without_docker_image_script(yt_path: str) -> None:
    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/dummy_script.py"],
        args=[
            "--mesh.node-count",
            "1",
            "--mesh.process-per-node",
            "1",
            "--mesh.gpu-per-process",
            "0",
            "--yt-path",
            yt_path,
            "--bind-local",
            f"{get_data_path('../data/dummy_script.py')}:/tractorun_tests/dummy_script.py",
        ],
        docker_image=None,
    )
    with pytest.raises(AssertionError):
        tracto_cli.dry_run()


def test_without_docker_image_pickle(yt_path: str) -> None:
    def checker() -> None:
        pass

    with pytest.raises(TractorunConfigurationError):
        run(
            checker,
            yt_path=yt_path,
            mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
            backend=GenericBackend(),
            dry_run=True,
        )


def test_run_cli_command_from_python(yt_path: str) -> None:
    run_script(
        ["python3", "/tractorun_tests/dummy_script.py"],
        yt_path=yt_path,
        binds_local=[
            BindLocal(
                source=str(get_data_path("../data/dummy_script.py")), destination="/tractorun_tests/dummy_script.py"
            ),
        ],
        binds_local_lib=[
            str(get_data_path("../../tractorun")),
        ],
        docker_image=GENERIC_DOCKER_IMAGE,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
    )


def test_script_debug_info(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/dummy_script.py"],
        args=[
            "--mesh.node-count",
            "1",
            "--mesh.process-per-node",
            "1",
            "--mesh.gpu-per-process",
            "0",
            "--yt-path",
            yt_path,
            "--bind-local",
            f"{get_data_path('../data/dummy_script.py')}:/tractorun_tests/dummy_script.py",
            "--no-wait",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    serializer = AttrSerializer(CliRunInfo)
    run_info = serializer.deserialize(op_run.stdout.decode("utf-8"))
    assert run_info.run_info is not None
    assert run_info.run_info.operation_id is not None


def test_change_working_dirs(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    def checker(toolbox: Toolbox) -> None:
        assert os.environ["HOME"] == "/tmp"
        assert os.environ["PWD"] == "/tmp"
        assert os.environ["TMPDIR"] == "/tmp"

    yt_client = yt_instance.get_client()

    operation_title = f"test operation {uuid.uuid4()}"
    task_title = f"test operation's task {uuid.uuid4()}"

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)

    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        env=[
            EnvVariable(
                name="HOME",
                value="/tmp",
            ),
            EnvVariable(
                name="PWD",
                value="/tmp",
            ),
            EnvVariable(
                name="TMPDIR",
                value="/tmp",
            ),
        ],
        docker_image=GENERIC_DOCKER_IMAGE,
        title=operation_title,
        yt_task_spec={"title": task_title},
    )



import enum
import io
import os
import subprocess
from typing import (
    TYPE_CHECKING,
    NewType,
)

import attrs

from tractorun.sidecar import (
    RestartPolicy,
    Sidecar,
)


SidecarIndex = NewType("SidecarIndex", int)


class RestartVerdict(enum.IntEnum):
    skip = enum.auto()
    restart = enum.auto()
    fail = enum.auto()
    unknown = enum.auto()


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class SidecarRun:
    _sidecar: Sidecar
    _process: subprocess.Popen
    _env: dict[str, str]

    @property
    def command(self) -> list[str]:
        return self._sidecar.command

    @classmethod
    def _run_process(cls, sidecar: Sidecar, env: dict[str, str]) -> subprocess.Popen:
        process = subprocess.Popen(
            sidecar.command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=1,
            universal_newlines=True,
            env=env,
            errors="replace",
        )
        if TYPE_CHECKING:
            assert isinstance(process.stdout, io.TextIOWrapper)
            assert isinstance(process.stderr, io.TextIOWrapper)
        os.set_blocking(process.stdout.fileno(), False)
        os.set_blocking(process.stderr.fileno(), False)
        return process

    @classmethod
    def run(cls, sidecar: Sidecar, env: dict[str, str]) -> "SidecarRun":
        process = cls._run_process(sidecar=sidecar, env=env)
        return SidecarRun(sidecar=sidecar, process=process, env=env)

    def poll(self) -> int | None:
        return self._process.poll()

    def wait(self) -> None:
        self._process.wait()

    def restart(self) -> "SidecarRun":
        self.terminate()
        return self.run(sidecar=self._sidecar, env=self._env)

    def terminate(self) -> None:
        self._process.terminate()

    def stdout(self) -> io.TextIOWrapper:
        if TYPE_CHECKING:
            assert isinstance(self._process.stdout, io.TextIOWrapper)
        return self._process.stdout

    def stderr(self) -> io.TextIOWrapper:
        if TYPE_CHECKING:
            assert isinstance(self._process.stderr, io.TextIOWrapper)
        return self._process.stderr

    def need_restart(self) -> RestartVerdict:
        exit_code = self._process.poll()
        if exit_code is None:
            return RestartVerdict.skip
        match self._sidecar.restart_policy:
            case RestartPolicy.ON_FAILURE:
                if exit_code == 0:
                    return RestartVerdict.skip
                return RestartVerdict.restart
            case RestartPolicy.ALWAYS:
                return RestartVerdict.restart
            case RestartPolicy.NEVER:
                return RestartVerdict.skip
            case RestartPolicy.FAIL:
                return RestartVerdict.fail
            case _:
                return RestartVerdict.unknown



import json
import os
from typing import (
    Generic,
    Type,
    TypeVar,
)

import attrs
import cattrs
import yt.wrapper as yt


_T = TypeVar("_T")


def create_attrs_converter(forbid_extra_keys: bool = True) -> cattrs.Converter:
    converter = cattrs.Converter()
    converter.register_structure_hook_factory(
        attrs.has,
        lambda cl: cattrs.gen.make_dict_structure_fn(cl, converter, _cattrs_forbid_extra_keys=forbid_extra_keys),
    )
    return converter


@attrs.define(slots=True)
class AttrSerializer(Generic[_T]):
    _type: Type[_T] = attrs.field()
    _converter: cattrs.Converter = attrs.field(factory=create_attrs_converter)

    def serialize(self, data: _T) -> str:
        return json.dumps(self._converter.unstructure(data))

    def deserialize(self, data: str) -> _T:
        return cattrs.structure(json.loads(data), self._type)


def create_prerequisite_client(yt_client: yt.YtClient, prerequisite_transaction_ids: list[str]) -> yt.YtClient:
    if yt_client:
        try:
            prerequisite_transaction_ids = prerequisite_transaction_ids + yt_client.get_option(
                "prerequisite_transaction_ids"
            )
        except Exception:
            pass
    return yt.create_client_with_command_params(yt_client, prerequisite_transaction_ids=prerequisite_transaction_ids)


def get_default_docker_image() -> str | None:
    # use the same env var as yt sdk
    return os.environ.get("YT_BASE_LAYER") or os.environ.get("YT_JOB_DOCKER_IMAGE") or None



from typing import (
    Any,
    Callable,
)
import warnings

import yt.wrapper as yt

from tractorun.base_backend import BackendBase
from tractorun.bind import (
    BindCypress,
    BindLocal,
)
from tractorun.docker_auth import DockerAuthData
from tractorun.env import EnvVariable
from tractorun.exception import TractorunConfigurationError
from tractorun.mesh import Mesh
from tractorun.operation_log import OperationLogMode
from tractorun.private.constants import DEFAULT_CLUSTER_CONFIG_PATH as _DEFAULT_CLUSTER_CONFIG_PATH
from tractorun.private.helpers import get_default_docker_image as _get_default_docker_image
from tractorun.private.run_internal import CliCommand as _CliCommand
from tractorun.private.run_internal import TractorunParams as _TractorunParams
from tractorun.private.run_internal import UserFunction as _UserFunction
from tractorun.private.run_internal import prepare_and_get_toolbox as _prepare_and_get_toolbox
from tractorun.private.run_internal import run_local as _run_local
from tractorun.private.run_internal import run_tracto as _run_tracto
from tractorun.resources import Resources
from tractorun.run_info import RunInfo
from tractorun.sidecar import Sidecar
from tractorun.stderr_reader import StderrMode
from tractorun.tensorproxy import TensorproxySidecar
from tractorun.toolbox import Toolbox


__all__ = ["run", "run_script", "prepare_and_get_toolbox"]


def run_script(
    user_command: list[str],
    *,
    yt_path: str,
    mesh: Mesh,
    title: str | None = None,
    user_config: dict[Any, Any] | None = None,
    cluster_config_path: str = _DEFAULT_CLUSTER_CONFIG_PATH,
    docker_image: str | None = None,
    resources: Resources | None = None,
    yt_client: yt.YtClient | None = None,
    binds_local: list[BindLocal] | None = None,
    binds_local_lib: list[str] | None = None,
    binds_cypress: list[BindCypress] | None = None,
    sidecars: list[Sidecar] | None = None,
    tensorproxy: TensorproxySidecar | None = None,
    env: list[EnvVariable] | None = None,
    no_wait: bool = False,
    yt_operation_spec: dict[Any, Any] | None = None,
    yt_task_spec: dict[Any, Any] | None = None,
    local: bool = False,
    proxy_stderr_mode: StderrMode = StderrMode.disabled,
    operation_log_mode: OperationLogMode = OperationLogMode.default,
    docker_auth: DockerAuthData | None = None,
    dry_run: bool = False,
) -> RunInfo:
    docker_image = _get_docker_image(docker_image)

    sidecars = sidecars or []
    binds_local = binds_local or []
    binds_local_lib = binds_local_lib or []
    binds_cypress = binds_cypress or []
    env = env or []
    user_config = user_config or {}
    resources = resources or Resources()
    yt_operation_spec = yt_operation_spec or {}
    yt_task_spec = yt_task_spec or {}

    params = _TractorunParams(
        runnable=_CliCommand(command=user_command),
        yt_path=yt_path,
        mesh=mesh,
        title=title,
        binds_local=binds_local,
        binds_local_lib=binds_local_lib,
        binds_cypress=binds_cypress,
        sidecars=sidecars,
        tensorproxy=tensorproxy,
        env=env,
        no_wait=no_wait,
        user_config=user_config,
        cluster_config_path=cluster_config_path,
        resources=resources,
        yt_client=yt_client,
        docker_image=docker_image,
        yt_operation_spec=yt_operation_spec,
        yt_task_spec=yt_task_spec,
        proxy_stderr_mode=proxy_stderr_mode,
        operation_log_mode=operation_log_mode,
        docker_auth=docker_auth,
        dry_run=dry_run,
        attach_external_libs=False,
    )
    if local:
        return _run_local(params=params)
    else:
        return _run_tracto(params=params)


def run(
    user_command: Callable,
    *,
    backend: BackendBase,
    yt_path: str,
    mesh: Mesh,
    title: str | None = None,
    user_config: dict[Any, Any] | None = None,
    cluster_config_path: str = _DEFAULT_CLUSTER_CONFIG_PATH,
    docker_image: str | None = None,
    resources: Resources | None = None,
    yt_client: yt.YtClient | None = None,
    binds_local: list[BindLocal] | None = None,
    binds_local_lib: list[str] | None = None,
    binds_cypress: list[BindCypress] | None = None,
    sidecars: list[Sidecar] | None = None,
    tensorproxy: TensorproxySidecar | None = None,
    env: list[EnvVariable] | None = None,
    no_wait: bool = False,
    yt_operation_spec: dict[Any, Any] | None = None,
    yt_task_spec: dict[Any, Any] | None = None,
    local: bool = False,
    proxy_stderr_mode: StderrMode = StderrMode.disabled,
    operation_log_mode: OperationLogMode = OperationLogMode.default,
    docker_auth: DockerAuthData | None = None,
    attach_external_libs: bool = False,
    dry_run: bool = False,
) -> RunInfo:
    if attach_external_libs:
        warnings.warn("Use attach_external_libs=True only in adhoc scripts. Don't use it in production.")
    docker_image = _get_docker_image(docker_image)

    sidecars = sidecars or []
    binds_local = binds_local or []
    binds_local_lib = binds_local_lib or []
    binds_cypress = binds_cypress or []
    env = env or []
    user_config = user_config or {}
    resources = resources or Resources()
    yt_operation_spec = yt_operation_spec or {}
    yt_task_spec = yt_task_spec or {}

    params = _TractorunParams(
        runnable=_UserFunction(function=user_command, backend=backend),
        yt_path=yt_path,
        mesh=mesh,
        title=title,
        binds_local=binds_local,
        binds_local_lib=binds_local_lib,
        binds_cypress=binds_cypress,
        sidecars=sidecars,
        tensorproxy=tensorproxy,
        env=env,
        no_wait=no_wait,
        user_config=user_config,
        cluster_config_path=cluster_config_path,
        resources=resources,
        yt_client=yt_client,
        docker_image=docker_image,
        yt_operation_spec=yt_operation_spec,
        yt_task_spec=yt_task_spec,
        proxy_stderr_mode=proxy_stderr_mode,
        operation_log_mode=operation_log_mode,
        docker_auth=docker_auth,
        attach_external_libs=attach_external_libs,
        dry_run=dry_run,
    )
    if local:
        return _run_local(params=params)
    else:
        return _run_tracto(params=params)


def _get_docker_image(docker_image: str | None) -> str:
    if docker_image is not None:
        return docker_image
    docker_image = _get_default_docker_image()
    if docker_image is None:
        raise TractorunConfigurationError("docker_image should be specified")
    return docker_image


def prepare_and_get_toolbox(backend: BackendBase) -> Toolbox:
    return _prepare_and_get_toolbox(backend)



import pathlib

from tractorun.backend.generic.backend import GenericBackend
from tractorun.run import prepare_and_get_toolbox


if __name__ == "__main__":
    toolbox = prepare_and_get_toolbox(backend=GenericBackend())
    user_config = toolbox.get_user_config()
    cypress_file_to_check = user_config["CYPRESS_FILE_TO_CHECK"]

    assert pathlib.Path(cypress_file_to_check).exists()
    assert pathlib.Path(cypress_file_to_check).is_file()



import enum

import attrs


__all__ = ["Sidecar", "RestartPolicy"]

from tractorun.private.helpers import AttrSerializer as _AttrSerializer


class RestartPolicy(str, enum.Enum):
    ON_FAILURE = "on_failure"
    ALWAYS = "always"
    NEVER = "never"
    FAIL = "fail"


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class Sidecar:
    command: list[str]
    restart_policy: RestartPolicy

    @staticmethod
    def from_args(value: str) -> "Sidecar":
        return _AttrSerializer(Sidecar).deserialize(value)



import sys

from jax import (
    grad,
    jit,
)

from tractorun.backend.tractorax import Tractorax
from tractorun.run import prepare_and_get_toolbox


def main() -> None:
    _ = prepare_and_get_toolbox(backend=Tractorax())

    @jit
    def f(x: int) -> int:
        return x**2 + 3 * x + 1

    grad_f = grad(f)
    print(grad_f(1.0), file=sys.stderr)


if __name__ == "__main__":
    main()



from tractorun.base_backend import EnvironmentBase
from tractorun.private.closet import Closet as _Closet


__all__ = ["Environment"]


class Environment(EnvironmentBase):
    def prepare(self, closet: _Closet) -> None:
        return None



import datetime
import sys
import time

import attrs
from yt import wrapper as yt
from yt.wrapper.errors import YtResolveError

from tractorun.coordinator import Coordinator
from tractorun.mesh import Mesh
from tractorun.private.helpers import create_prerequisite_client
from tractorun.private.training_dir import TrainingDir
from tractorun.private.training_dir import TrainingDir as _TrainingDir


def get_incarnation_id(yt_client: yt.YtClient, training_dir: TrainingDir, raise_if_not_exists: bool = False) -> int:
    try:
        return yt_client.get(training_dir.base_path + "/@incarnation_id")
    except YtResolveError:
        if raise_if_not_exists:
            raise
        return -1


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class CoordinatorFactory:
    _self_endpoint: str
    _self_index: int
    _process_index: int
    _node_index: int
    _mesh: Mesh
    _yt_client: yt.YtClient
    _training_dir: _TrainingDir
    _operation_id: str
    _job_id: str

    def create(self) -> Coordinator:
        if self._self_index == 0:
            return self._make_primary(self_index=self._self_index)
        else:
            return self._make_subordinate(self_index=self._self_index)

    def _wait_for_gang_barrier(self, incarnation_path: str) -> None:
        print("Waiting for all peers to start", file=sys.stderr)
        while True:
            try:
                topology = self._yt_client.get(incarnation_path + "/@topology")
                if all(peer["address"] != "" for peer in topology):
                    print("All peers started", file=sys.stderr)
                    break
            except Exception as e:
                print(f"_wait_for_gang_barrier raised exception {e}", file=sys.stderr)
            time.sleep(1.0)

    def _make_primary(self, self_index: int) -> "Coordinator":
        incarnation_transaction_id = self._yt_client.start_transaction()
        assert incarnation_transaction_id is not None

        with self._yt_client.Transaction(
            transaction_id=incarnation_transaction_id,
            acquire=False,
            ping=False,
        ):
            self._yt_client.lock(
                self._training_dir.primary_lock_path,
                mode="exclusive",
                waitable=True,
                wait_for=int(datetime.timedelta(minutes=5).total_seconds() * 1000),
            )

        last_incarnation_id = get_incarnation_id(self._yt_client, self._training_dir)
        incarnation_id = last_incarnation_id + 1

        incarnation_yt_client = create_prerequisite_client(
            self._yt_client,
            [incarnation_transaction_id],
        )

        with incarnation_yt_client.Transaction():
            incarnation_yt_client.set(
                self._training_dir.base_path + "/@incarnation_id",
                incarnation_id,
            )

        with incarnation_yt_client.Transaction():
            incarnation_path = self._training_dir.get_incarnation_path(incarnation_id)
            incarnation_yt_client.create("map_node", incarnation_path)
            incarnation_yt_client.set(
                incarnation_path + "/@incarnation_transaction_id",
                incarnation_transaction_id,
            )
            incarnation_yt_client.set(
                incarnation_path + "/@incarnation_operation_id",
                self._operation_id,
            )
            incarnation_yt_client.set(
                incarnation_path + "/@primary_endpoint",
                self._self_endpoint,
            )

            topology = [
                {
                    "address": self._self_endpoint,
                    "job_id": self._job_id,
                },
            ] + [
                {"address": "", "job_id": ""}
            ] * (self._mesh.peer_count - 1)
            incarnation_yt_client.set(
                incarnation_path + "/@topology",
                topology,
            )

        self._wait_for_gang_barrier(incarnation_path)

        return Coordinator(
            self_index=self_index,
            incarnation_id=incarnation_id,
            mesh=self._mesh,
            process_index=self._process_index,
            self_endpoint=self._self_endpoint,
            primary_endpoint=self._self_endpoint,
        )

    def _make_subordinate(self, self_index: int) -> Coordinator:
        while True:
            try:
                incarnation_id = get_incarnation_id(
                    self._yt_client,
                    self._training_dir,
                    raise_if_not_exists=True,
                )
                incarnation_path = self._training_dir.get_incarnation_path(incarnation_id)
                if (
                    self._yt_client.get(
                        incarnation_path + "/@incarnation_operation_id",
                    )
                    != self._operation_id
                ):
                    raise RuntimeError("Operation id mismatch")

                # incarnation_transaction_id: str = self._yt_client.get(
                #     incarnation_path + "/@incarnation_transaction_id",
                # )
                # assert incarnation_transaction_id is not None
                # incarnation_yt_client = create_prerequisite_client(
                #     self._yt_client,
                #     [incarnation_transaction_id],
                # )
                incarnation_yt_client = self._yt_client

                incarnation_yt_client.set(
                    incarnation_path + f"/@topology/{self_index}",
                    {"address": self._self_endpoint, "job_id": self._job_id},
                )

                primary_endpoint = incarnation_yt_client.get(
                    incarnation_path + "/@primary_endpoint",
                )

                self._wait_for_gang_barrier(incarnation_path)
            except Exception:
                time.sleep(1.0)
                continue

            return Coordinator(
                self_index=self_index,
                incarnation_id=incarnation_id,
                mesh=self._mesh,
                process_index=self._process_index,
                self_endpoint=self._self_endpoint,
                primary_endpoint=primary_endpoint,
            )



import json
import os
import pathlib
import typing as t

import pytest
import yt.wrapper as yt

from tests.utils import (
    GENERIC_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.bind import (
    BindAttributes,
    BindCypress,
    BindLocal,
)
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.run import run
from tractorun.toolbox import Toolbox


FILE_PATH = "/bind/file"
DIR_PATH = "/bind/dir"


@pytest.fixture
def cypress_file(yt_instance: YtInstance, yt_path: str) -> str:
    yt_client = yt_instance.get_client()
    file_path = f"{yt_path}/cypress_file"
    yt_client.write_file(file_path, b"hello")
    yt_client.set_attribute(file_path, "executable", True)
    return file_path


def test_configuration_cypress() -> None:
    _, _, config = make_configuration(
        make_cli_args("--bind-cypress", "//tmp/cli:cli", "command"),
    )
    assert config.bind_cypress == [BindCypress(source="//tmp/cli", destination="cli")]

    _, _, config = make_configuration(
        make_cli_args(
            "--bind-cypress",
            json.dumps(
                {
                    "source": "//tmp/cli",
                    "destination": "cli",
                }
            ),
            "command",
        ),
    )
    assert config.bind_cypress == [BindCypress(source="//tmp/cli", destination="cli", attributes=BindAttributes())]

    _, _, config = make_configuration(
        make_cli_args(
            "--bind-cypress",
            json.dumps(
                {
                    "source": "//tmp/cli",
                    "destination": "cli",
                    "attributes": {"format": "cli"},
                }
            ),
            "command",
        ),
    )
    assert config.bind_cypress == [
        BindCypress(source="//tmp/cli", destination="cli", attributes=BindAttributes(format="cli"))
    ]

    run_config = make_run_config(
        {
            "bind_cypress": [
                {
                    "source": "//tmp/config",
                    "destination": "config",
                    "attributes": {"format": "config"},
                }
            ],
        },
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.bind_cypress == [
        BindCypress(source="//tmp/config", destination="config", attributes=BindAttributes(format="config"))
    ]

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            make_cli_args(
                "--bind-cypress",
                json.dumps(
                    {
                        "source": "//tmp/cli",
                        "destination": "cli",
                        "attributes": {"format": "cli"},
                    }
                ),
                "command",
            ),
        )
    assert config.bind_cypress == [
        BindCypress(source="//tmp/cli", destination="cli", attributes=BindAttributes(format="cli"))
    ]


def test_configuration_local() -> None:
    _, _, config = make_configuration(make_cli_args("--bind-local", "//tmp/cli:/tmp/cli"))
    assert config.bind_local == [BindLocal(source="//tmp/cli", destination="/tmp/cli")]

    _, _, config = make_configuration(
        make_cli_args("--bind-cypress", json.dumps({"source": "//tmp/cli", "destination": "/tmp/cli"}), "command"),
    )
    assert config.bind_cypress == [BindCypress(source="//tmp/cli", destination="/tmp/cli")]

    run_config = make_run_config(
        {
            "bind_local": [
                {
                    "source": "//tmp/config",
                    "destination": "/tmp/config",
                }
            ],
        }
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.bind_local == [BindLocal(source="//tmp/config", destination="/tmp/config")]

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path, "--bind-local", "//tmp/cli:/tmp/cli"])
    assert config.bind_local == [BindLocal(source="//tmp/cli", destination="/tmp/cli")]


def checker_dir(toolbox: Toolbox) -> None:
    assert pathlib.Path(DIR_PATH).is_dir()
    assert (pathlib.Path(DIR_PATH) / "some_file").is_file()


def get_file_checker(
    file_paths: list[str] | None = None, check_executable: bool = False
) -> t.Callable[[Toolbox], None]:
    if file_paths is None:
        file_paths = [FILE_PATH]

    def checker_file(toolbox: Toolbox) -> None:
        for file_path in file_paths:
            assert pathlib.Path(file_path).is_file()
            if check_executable:
                assert os.access(file_path, os.X_OK)

    return checker_file


def test_local_bind_file_pickle(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    run(
        get_file_checker(),
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_local=[
            BindLocal(
                source=str(get_data_path("../data/binds/another_file")),
                destination=FILE_PATH,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
    )


def test_local_bind_dir_pickle(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    run(
        checker_dir,
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_local=[
            BindLocal(
                source=str(get_data_path("../data/binds/bind_dir")),
                destination=DIR_PATH,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
    )


def test_cypress_bind_file(yt_instance: YtInstance, yt_path: str, cypress_file: str) -> None:
    yt_client = yt_instance.get_client()

    destination_path = "bar"

    run(
        get_file_checker([destination_path], check_executable=True),
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_cypress=[
            BindCypress(
                source=cypress_file,
                destination=destination_path,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
    )


def test_cypress_bind_file_attrs(yt_instance: YtInstance, yt_path: str, cypress_file: str) -> None:
    yt_client = yt_instance.get_client()

    def checker() -> None:
        pass

    run_info = run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
        dry_run=True,
        binds_cypress=[
            BindCypress(
                source=cypress_file,
                destination="local_file",
                attributes=BindAttributes(
                    executable=False,
                    format="dummy",
                    bypass_artifact_cache=True,
                ),
            )
        ],
    )
    attached_file = run_info.operation_spec["tasks"]["task"]["file_paths"][-1]
    assert str(attached_file) == cypress_file
    assert attached_file.attributes == {
        "executable": False,
        "file_name": "local_file",
        "bypass_artifact_cache": True,
        "format": "dummy",
    }


def test_cypress_bind_map_node(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    yt_map_path = f"{yt_path}/some_dir"
    yt_client.create("map_node", yt_map_path)
    yt_client.create("map_node", f"{yt_map_path}/nested")

    destination = "some_folder"

    file_suffixes = ["foo", "bar", "nested/some", "nested/another"]
    for file_suffix in file_suffixes:
        file_path = f"{yt_map_path}/{file_suffix}"
        yt_client.write_file(file_path, b"hello")
        yt_client.set_attribute(file_path, "executable", True)

    run(
        get_file_checker(
            file_paths=[f"{destination}/{file_suffix}" for file_suffix in file_suffixes],
            check_executable=True,
        ),
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_cypress=[
            BindCypress(
                source=yt_map_path,
                destination=destination,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
    )


def test_cypress_bind_map_node_spec(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    yt_map_path = f"{yt_path}/some_dir"
    yt_client.create("map_node", yt_map_path)
    yt_client.create("map_node", f"{yt_map_path}/nested")

    # test warning for non-file nodes
    yt_client.create("document", f"{yt_map_path}/document")

    destination = "some_local_dir"

    file_suffixes = ["foo", "bar", "nested/some", "nested/another"]
    for file_suffix in file_suffixes:
        file_path = f"{yt_map_path}/{file_suffix}"
        yt_client.write_file(file_path, b"hello")
        yt_client.set_attribute(file_path, "executable", True)

    with pytest.warns(UserWarning, match="Skip .*/document because it is not a file, but document"):
        run_info = run(
            get_file_checker(file_suffixes, check_executable=True),
            backend=GenericBackend(),
            yt_path=yt_path,
            binds_cypress=[
                BindCypress(
                    source=f"{yt_map_path}",
                    destination=destination,
                    attributes=BindAttributes(
                        executable=False,
                        format="dummy",
                        bypass_artifact_cache=True,
                    ),
                ),
            ],
            mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
            yt_client=yt_client,
            docker_image=GENERIC_DOCKER_IMAGE,
            dry_run=True,
        )
    attached_files = run_info.operation_spec["tasks"]["task"]["file_paths"]
    attached_files = [
        attach for attach in attached_files if isinstance(attach, yt.ypath.FilePath)  # skip yt wrapper files
    ]
    attached_files = list(sorted(attached_files, key=str))
    file_suffixes = list(sorted(file_suffixes))
    for file_suffix, attached_file in zip(file_suffixes, attached_files):
        assert str(attached_file) == f"{yt_map_path}/{file_suffix}"
        assert attached_file.attributes == {
            "executable": False,
            "file_name": f"{destination}/{file_suffix}",
            "bypass_artifact_cache": True,
            "format": "dummy",
        }


def test_cypress_bind_map_node_symlinks(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    yt_map_path = f"{yt_path}/some_dir"
    yt_map_path_1 = f"{yt_map_path}/1"
    yt_map_path_2 = f"{yt_map_path}/2"

    yt_client.create("map_node", yt_map_path)
    yt_client.create("map_node", yt_map_path_1)
    yt_client.create("map_node", yt_map_path_2)
    yt_client.link(link_path=f"{yt_map_path_1}/link_to_2", target_path=yt_map_path_2)

    destination = "some_local_dir"

    file_suffixes = ["foo", "bar"]
    for base_folder in [yt_map_path_1, yt_map_path_2]:
        for file_suffix in file_suffixes:
            file_path = f"{base_folder}/{file_suffix}"
            yt_client.write_file(file_path, b"hello")

    def checker(_: Toolbox) -> None:
        pass

    run_info = run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_cypress=[
            BindCypress(
                source=yt_map_path_1,
                destination=destination,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
        dry_run=True,
    )
    attached_files = run_info.operation_spec["tasks"]["task"]["file_paths"]
    attached_files = [
        str(attach) for attach in attached_files if isinstance(attach, yt.ypath.FilePath)  # skip yt wrapper files
    ]
    attached_files = list(sorted(attached_files))
    assert attached_files == [
        f"{yt_path}/some_dir/1/bar",
        f"{yt_path}/some_dir/1/foo",
        f"{yt_path}/some_dir/1/link_to_2/bar",
        f"{yt_path}/some_dir/1/link_to_2/foo",
    ]


def test_cypress_bind_from_run_config(yt_instance: YtInstance, yt_path: str, cypress_file: str) -> None:
    yt_client = yt_instance.get_client()

    run_config = {
        "mesh": {
            "node_count": 1,
            "process_per_node": 1,
            "gpu_per_process": 0,
        },
        "bind_cypress": [
            {
                "source": f"{cypress_file}",
                "destination": "bar",
            },
        ],
    }

    with run_config_file(run_config) as run_config_path:
        tracto_cli = TractoCli(
            command=["python3", "/tractorun_tests/check_cypress_bind.py"],
            args=[
                "--run-config-path",
                run_config_path,
                "--yt-path",
                yt_path,
                "--bind-local",
                f"{get_data_path('../data/check_cypress_bind.py')}:/tractorun_tests/check_cypress_bind.py",
                "--user-config",
                json.dumps({"CYPRESS_FILE_TO_CHECK": "./bar"}),
            ],
        )
        op_run = tracto_cli.run()

    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)



import os

from tractorun.backend.generic import GenericBackend
from tractorun.run import prepare_and_get_toolbox


def main() -> None:
    toolbox = prepare_and_get_toolbox(backend=GenericBackend())
    user_config = toolbox.get_user_config()
    assert os.environ["SECRET"] == user_config["secret_env_value"]
    assert os.environ["NOT_SECRET"] == user_config["not_secret_env_value"]


if __name__ == "__main__":
    main()



import json
import os

import pytest
import yt.wrapper as yt

from tests.utils import (
    GENERIC_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.env import EnvVariable
from tractorun.mesh import Mesh
from tractorun.run import run


SECRET_ENV_VALUE = "secret"
NOT_SECRET_ENV_VALUE = "not_secret"


def test_configuration() -> None:
    _, _, config = make_configuration(make_cli_args())
    assert config.env == []

    env_json = json.dumps(
        {
            "name": "cli",
            "cypress_path": "raw_cypress_path",
            "value": "raw_value",
        }
    )

    _, _, config = make_configuration(
        make_cli_args(
            "--env",
            env_json,
        ),
    )
    assert config.env == [EnvVariable(name="cli", value="raw_value", cypress_path="raw_cypress_path")]

    run_config = make_run_config(
        {
            "env": [
                {
                    "name": "config",
                    "cypress_path": "raw_cypress_path",
                    "value": "raw_value",
                },
            ]
        },
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.env == [EnvVariable(name="config", value="raw_value", cypress_path="raw_cypress_path")]

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            ["--run-config-path", run_config_path, "--env", env_json],
        )
    assert config.env == [EnvVariable(name="cli", value="raw_value", cypress_path="raw_cypress_path")]


@pytest.fixture
def yt_secret_path(yt_instance: YtInstance, yt_path: str) -> str:
    yt_client = yt_instance.get_client()

    secret_path = yt_path + "/secret"
    yt.create("document", secret_path, client=yt_client)
    yt.set(secret_path, SECRET_ENV_VALUE, client=yt_client)
    return secret_path


def test_environment(yt_secret_path: str, yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    def env_checker(toolbox: TractoCli) -> None:
        assert os.environ["SECRET"] == SECRET_ENV_VALUE
        assert os.environ["NOT_SECRET"] == NOT_SECRET_ENV_VALUE

    run(
        env_checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
        env=[
            EnvVariable(name="SECRET", cypress_path=yt_secret_path),
            EnvVariable(name="NOT_SECRET", value="not_secret"),
        ],
    )


def test_run_script(yt_instance: YtInstance, yt_secret_path: str, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/env_script.py"],
        args=[
            "--yt-path",
            yt_path,
            "--env",
            json.dumps(
                {
                    "name": "SECRET",
                    "cypress_path": yt_secret_path,
                },
            ),
            "--env",
            json.dumps(
                {
                    "name": "NOT_SECRET",
                    "value": NOT_SECRET_ENV_VALUE,
                },
            ),
            "--user-config",
            json.dumps({"secret_env_value": SECRET_ENV_VALUE, "not_secret_env_value": NOT_SECRET_ENV_VALUE}),
            "--bind-local",
            f"{get_data_path('../data/env_script.py')}:/tractorun_tests/env_script.py",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)


def test_run_script_with_config(yt_instance: YtInstance, yt_path: str, yt_secret_path: str) -> None:
    yt_client = yt_instance.get_client()

    run_config = {
        "mesh": {
            "node_count": 1,
            "process_per_node": 1,
            "gpu_per_process": 0,
        },
        "env": [
            {
                "name": "SECRET",
                "cypress_path": yt_secret_path,
            },
            {
                "name": "NOT_SECRET",
                "value": NOT_SECRET_ENV_VALUE,
            },
        ],
        "user_config": {
            "secret_env_value": SECRET_ENV_VALUE,
            "not_secret_env_value": NOT_SECRET_ENV_VALUE,
        },
    }
    with run_config_file(run_config) as run_config_path:
        tracto_cli = TractoCli(
            command=["python3", "/tractorun_tests/env_script.py"],
            args=[
                "--run-config-path",
                run_config_path,
                "--yt-path",
                yt_path,
                "--bind-local",
                f"{get_data_path('../data/env_script.py')}:/tractorun_tests/env_script.py",
            ],
        )
        op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)



import asyncio
import copy
from functools import partial
import json
from typing import (
    Generic,
    TypeVar,
)

import attrs
import yt.wrapper as yt


_T = TypeVar("_T")


__all__ = ["CheckpointManager", "Checkpoint"]


@attrs.define
class Checkpoint:
    index: int
    value: bytes
    metadata: dict


def _save_checkpoint(yt_client: yt.YtClient, path: str, metadata: bytes, value: bytes) -> None:
    with yt_client.Transaction():
        yt_client.create("map_node", path)
        yt_client.write_file(path + "/value", value)
        yt_client.write_file(
            path + "/metadata",
            metadata,
        )


@attrs.define
class _Task(Generic[_T]):
    _task: asyncio.Future[_T]

    def wait(self, timeout: int) -> _T:
        return asyncio.get_event_loop().run_until_complete(asyncio.wait_for(self._task, timeout=timeout))


@attrs.define
class CheckpointManager:
    _path: str
    _yt_client: yt.YtClient
    _last_checkpoint_index: int = -1

    def initialize(self) -> None:
        last_checkpoint_index = -1
        for index in self._yt_client.list(self._path):
            try:
                index = int(index)
            except Exception:
                continue
            last_checkpoint_index = max(last_checkpoint_index, index)
        self._last_checkpoint_index = last_checkpoint_index

    def get_last_checkpoint(self) -> Checkpoint | None:
        if self._last_checkpoint_index == -1:
            return None

        checkpoint_path = self._path + "/" + str(self._last_checkpoint_index)
        value = self._yt_client.read_file(checkpoint_path + "/value").read()
        metadata = json.loads(self._yt_client.read_file(checkpoint_path + "/metadata").read())

        return Checkpoint(self._last_checkpoint_index, value, metadata)

    def save_checkpoint(self, value: bytes, metadata: dict | None = None) -> _Task:
        if metadata is None:
            metadata = {}
        # TODO: prerequisites

        checkpoint_index = self._last_checkpoint_index + 1
        checkpoint_path = self._path + "/" + str(checkpoint_index)
        serialized_metadata = json.dumps(metadata).encode("utf-8")
        self._last_checkpoint_index = checkpoint_index

        yt_client = yt.YtClient(config=copy.deepcopy(self._yt_client.config))
        save_checkpoint_task = partial(
            _save_checkpoint,
            yt_client=yt_client,
            path=checkpoint_path,
            metadata=serialized_metadata,
            value=value,
        )
        task = asyncio.get_event_loop().run_in_executor(None, save_checkpoint_task)
        return _Task[None](
            task=task,
        )



import json
import time

from _pytest.capture import CaptureFixture
import pytest
import yt.wrapper as yt

from tests.utils import (
    GENERIC_DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.bind import BindLocal
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.private.sidecar import (
    RestartVerdict,
    SidecarRun,
)
from tractorun.run import run
from tractorun.sidecar import (
    RestartPolicy,
    Sidecar,
)
from tractorun.stderr_reader import StderrMode
from tractorun.toolbox import Toolbox


def test_configuration() -> None:
    _, _, config = make_configuration(make_cli_args())
    assert config.sidecars == []

    sidecars_json = json.dumps(
        {
            "command": ["cli"],
            "restart_policy": RestartPolicy.ON_FAILURE.value,
        },
    )

    _, _, config = make_configuration(
        make_cli_args(
            "--sidecar",
            sidecars_json,
        ),
    )
    assert config.sidecars == [Sidecar(command=["cli"], restart_policy=RestartPolicy.ON_FAILURE)]

    run_config = make_run_config(
        {
            "sidecars": [
                {
                    "command": ["config"],
                    "restart_policy": RestartPolicy.ON_FAILURE.value,
                }
            ]
        },
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.sidecars == [Sidecar(command=["config"], restart_policy=RestartPolicy.ON_FAILURE)]

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            ["--run-config-path", run_config_path, "--sidecar", sidecars_json],
        )
    assert config.sidecars == [Sidecar(command=["cli"], restart_policy=RestartPolicy.ON_FAILURE)]


def test_pickle(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()
    attr_key = "test_key"
    attr_value = "test_value"

    def checker(toolbox: Toolbox) -> None:
        import time

        client = toolbox.yt_client
        value = None
        attempts = 0
        while value is None:
            try:
                value = client.get(f"{yt_path}/@{attr_key}")
            except yt.errors.YtResolveError:
                pass
            time.sleep(5)
            if attempts > 5:
                raise Exception("Something wrong with sidecar")
            attempts += 1
        assert value == attr_value

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        sidecars=[
            Sidecar(
                command=["yt", "set", f"{yt_path}/@{attr_key}", attr_value],
                restart_policy=RestartPolicy.ON_FAILURE,
            )
        ],
        mesh=mesh,
        yt_client=yt_client,
        docker_image=GENERIC_DOCKER_IMAGE,
    )


def test_cli_args(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    attr_key = "test_key"
    attr_value = "test_value"

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/sidecar_script.py"],
        args=[
            "--yt-path",
            yt_path,
            "--user-config",
            json.dumps(
                {
                    "attr_key": attr_key,
                    "attr_value": attr_value,
                    "yt_path": yt_path,
                },
            ),
            "--sidecar",
            json.dumps(
                {
                    "command": ["yt", "set", f"{yt_path}/@{attr_key}", attr_value],
                    "restart_policy": RestartPolicy.ON_FAILURE,
                },
            ),
            "--bind-local",
            f"{get_data_path('../data/sidecar_script.py')}:/tractorun_tests/sidecar_script.py",
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)


def test_cli_config(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    attr_key = "test_key"
    attr_value = "test_value"

    run_config = {
        "mesh": {
            "node_count": 1,
            "process_per_node": 1,
            "gpu_per_process": 0,
        },
        "sidecars": [
            {
                "command": ["yt", "set", f"{yt_path}/@{attr_key}", attr_value],
                "restart_policy": RestartPolicy.ON_FAILURE.value,
            },
        ],
    }
    with run_config_file(run_config) as run_config_path:
        tracto_cli = TractoCli(
            command=["python3", "/tractorun_tests/sidecar_script.py"],
            args=[
                "--run-config-path",
                run_config_path,
                "--yt-path",
                yt_path,
                "--user-config",
                json.dumps(
                    {
                        "attr_key": attr_key,
                        "attr_value": attr_value,
                        "yt_path": yt_path,
                    },
                ),
                "--bind-local",
                f"{get_data_path('../data/sidecar_script.py')}:/tractorun_tests/sidecar_script.py",
            ],
        )
        op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)


COMMAND_SUCCESS = ["python3", "-c", "import sys; sys.exit(0)"]
COMMAND_FAILED = ["python3", "-c", "import sys; sys.exit(1)"]


@pytest.mark.parametrize(
    "command,policy,verdict",
    [
        (
            COMMAND_SUCCESS,
            RestartPolicy.ALWAYS,
            RestartVerdict.restart,
        ),
        (
            COMMAND_SUCCESS,
            RestartPolicy.ON_FAILURE,
            RestartVerdict.skip,
        ),
        (
            COMMAND_SUCCESS,
            RestartPolicy.FAIL,
            RestartVerdict.fail,
        ),
        (
            COMMAND_SUCCESS,
            RestartPolicy.NEVER,
            RestartVerdict.skip,
        ),
        (
            COMMAND_FAILED,
            RestartPolicy.ALWAYS,
            RestartVerdict.restart,
        ),
        (
            COMMAND_FAILED,
            RestartPolicy.ON_FAILURE,
            RestartVerdict.restart,
        ),
        (
            COMMAND_FAILED,
            RestartPolicy.FAIL,
            RestartVerdict.fail,
        ),
        (
            COMMAND_FAILED,
            RestartPolicy.NEVER,
            RestartVerdict.skip,
        ),
    ],
)
def test_need_restart(command: list[str], policy: RestartPolicy, verdict: RestartVerdict) -> None:
    sidecar = Sidecar(command=command, restart_policy=policy)
    sidecar_run = SidecarRun.run(
        sidecar=sidecar,
        env={},
    )
    sidecar_run.wait()
    assert sidecar_run.need_restart() == verdict


def test_sidecar_run() -> None:
    sidecar = Sidecar(command=["sleep", "infinity"], restart_policy=RestartPolicy.NEVER)
    sidecar_run = SidecarRun.run(sidecar=sidecar, env={})
    sidecar_run = sidecar_run.restart()
    assert sidecar_run.poll() is None
    sidecar_run.terminate()

    sidecar = Sidecar(command=["echo", "1"], restart_policy=RestartPolicy.NEVER)
    sidecar_run = SidecarRun.run(sidecar=sidecar, env={})
    sidecar_run.wait()
    sidecar_run = sidecar_run.restart()
    sidecar_run.terminate()


def test_restarting_sidecar_logs(yt_path: str, yt_instance: YtInstance, capsys: CaptureFixture[str]) -> None:
    yt_client = yt_instance.get_client()

    attr_key = "test_key"
    path = f"{yt_path}/@{attr_key}"

    yt_client.set(path, 0)

    def checker(toolbox: Toolbox) -> None:
        client = toolbox.yt_client
        value = 0
        while value < 10:
            value = client.get(path)
            print(f"main process read {value}")
            time.sleep(0.5)

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        sidecars=[
            Sidecar(
                command=["python3", "/tractorun_tests/restarting_sidecar_script.py", path],
                restart_policy=RestartPolicy.ALWAYS,
            ),
        ],
        binds_local=[
            BindLocal(
                source=f"{get_data_path('../data/restarting_sidecar_script.py')}",
                destination="/tractorun_tests/restarting_sidecar_script.py",
            ),
        ],
        docker_image=GENERIC_DOCKER_IMAGE,
        proxy_stderr_mode=StderrMode.primary,
    )
    captured = capsys.readouterr()
    for i in range(10):
        assert f"sidecar reads {i}" in captured.out



import importlib
import pkgutil
from types import ModuleType
from typing import Iterable

from tests.utils import check_no_private_objects_in_public_namespace


def _get_all_modules() -> Iterable[ModuleType]:
    import tractorun

    for module_info in pkgutil.walk_packages(tractorun.__path__, tractorun.__name__ + "."):
        # should base tractorun logic and  generic backend
        if module_info.name.startswith("tractorun.backend") and not module_info.name.startswith(
            "tractorun.backend.generic"
        ):
            continue
        yield importlib.import_module(module_info.name)


def test_no_private_objects_in_public_namespace() -> None:
    assert check_no_private_objects_in_public_namespace(_get_all_modules())
